import asyncio
from collections.abc import Generator
from typing import List

import aiofiles
from aiopath import AsyncPath
from asyncstdlib.functools import lru_cache

from .filesystems import FSHandlers
from .logs import logger


@lru_cache(maxsize=4096)
async def find_mount(path: AsyncPath) -> AsyncPath:
    """Find the mount point for a given path."""
    path = await path.absolute()
    while not await path.is_mount():
        path = path.parent
    return path


async def get_all_mounts() -> Generator:
    """Get a list of all mounted filesystems."""
    async with aiofiles.open("/proc/mounts", mode="r") as output:
        async for line in output:
            yield line.split()[1]


async def mount_bound_rglob(path: AsyncPath, mount: AsyncPath, pattern: str, ignoredirs: List[AsyncPath]) -> Generator:
    """Recursively glob a path, but stop at mount points."""
    path = await path.absolute()
    if str(path).startswith("/home/parallels"):
        logger.warning(f"Processing {str(path)}")
    path_mount = await find_mount(path)
    if path_mount != mount:
        logger.info(f"Skipping due to mount diff: {path} (wanted {mount})")
        return
    # Skip any ignored directories/files with a sanity check
    for ignore in ignoredirs:
        if str(path).startswith(str(ignore)) and await path.is_relative_to(ignore):
            logger.info(f"Skipping ignored subdir: {path}")
            return
    if await path.is_dir():
        async for subpath in path.glob(pattern):
            async for subitem in mount_bound_rglob(subpath, mount, pattern, ignoredirs):
                yield subitem
    yield path


@lru_cache(maxsize=1024)
async def mount_to_fs_handler(path: AsyncPath) -> str:
    # TODO: This is a hacky way to get the filesystem type, but it works for now.
    # Maybe with libblkid Python bindings?
    logger.info(f"Getting filesystem for mount: {path}")
    proc = await asyncio.create_subprocess_exec(
        "stat", "-f", "-L", "-c", "%T", str(path), stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE
    )
    stdout, _ = await proc.communicate()

    if proc.returncode != 0:
        err = f"Unable to get filesystem for {path}"
        logger.error(err)
        raise RuntimeError(err)

    fs = stdout.decode().strip()
    logger.info(f"Filesystem for {path}: {fs}")

    try:
        return FSHandlers[fs]
    except KeyError:
        err = f"Unsupported filesystem: {fs}"
        logger.error(err)
        raise RuntimeError(err)
